/*
    This file is part of the oolaboola digital dj program

    Copyright (C) 1999  Eric S. Tiedemann

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

extern "C"
{
#include <stddef.h>
}

class ButterJam
{
 public:

  enum type {NONE, LOWPASS, BANDPASS, HIGHPASS};

 private:

  struct filter
  {
    enum type t;

    float a[3];
    float b[2];
  
    float x[2];
    float y[2];

    float level;

    filter()
      : t(NONE), level(0)
      {
	x[0] = x[1] = 0.0;
	y[0] = y[1] = 0.0;
      }

    void adjust(float sr, enum type t, float f, float q=0);
  };

  // nchannels * nfilters dynamic array of filters
  filter **filters;

 public:

  const size_t nchannels, nfilters;
  const float sr;

  ButterJam(size_t nchannels, size_t nfilters, float sr);
  void adjust(size_t channel, size_t filter, float level, 
	      enum type t, float f, float q=0);
  void adjust(size_t channel, size_t filter, float level);
  void go(float *in, float *out, size_t nsamps);

};
