/*
    This file is part of the oolaboola digital dj program

    Copyright (C) 1999  Eric S. Tiedemann

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*

control msgs:
(output-to-file <file>)
(append-output-to-file <file>)
(end-file-output)
(file0 <file>)
(file1 <file>)
(rate0 <rate>)
(rate1 <rate>)
(position0 <position>)
(position1 <position>)
(position0-get)
(position1-get)
(eq0 <low level> <mid level> <high level>)
(eq1 <low level> <mid level> <high level>)
(echo0 <delay> <depth>)
(echo1 <delay> <depth>)
(energy0-get)
(energy1-get)
(position0-delta <delta>)
(position1-delta <delta>)
(setloop0 <start> <end>)
(clearloop0)
(setloop1 <start> <end>)
(clearloop1)
(cue0)
(cue1)
(level0 <rate>)
(level1 <rate>)
(xfade <share0> <share1>)
(status)
(shutdown)

*/

#include <cstdlib>
#include <exception>
#include <cstddef>
#include <new>
  extern "C" {
#include <unistd.h>
// for snprintf()
#define __USE_BSD
#include <stdio.h>
#include <sched.h>
#include <math.h>
#include <sys/mman.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <pthread.h>
	     }
#include <string>
#include <queue>
#include <iostream>

using namespace std;

#include "wavfilea.h"

//#define MUTEX_DEBUG
#include "mutex.h"
#include "soundout.h"
#include "sexpr.h"
#include "dprintf.h"
#ifdef OUTSERVICE
#include "ServiceQueue.h"
#endif
#include "ButterJam.h"

using namespace estlib;


const char progname[] = "mixd";

// global mixd mutex
pthread_mutex_t the_mutex = PTHREAD_MUTEX_INITIALIZER;

class EndApp : public exception {};

#if 0
void
dmpmutex(pthread_mutex_t *m)
{
  fprintf(stderr, "mutex %p: %d, %p, %p\n", 
	  m, m->m_count, m->m_waiting.head, m->m_waiting.tail);
}
#else
#define dmpmutex(m)
#endif

#ifdef OUTSERVICE
class outservice: public ServiceQueue<string>
{
public:

  void service(string s) 
    { 
      fputs(s.c_str(), stdout); 
      fflush(stdout);
    }

  outservice()
    : ServiceQueue<string>(ServiceQueue<string>::INHERIT_SCHEDULING)
    {}
};

outservice *os;
#else
queue<string> outq;
pthread_mutex_t outm;
pthread_cond_t outc, outempty;
pthread_t outt;

void *
outtask(void *v)
{
  mutex x(&outm);

  while (true)
    {
      while (outq.size() == 0)
	pthread_cond_wait(&outc, &outm);

      while (outq.size() > 0)
	{
	  string e = outq.front();
	  outq.pop();
	  {
	    nomutex x(&outm);
	    fputs(e.c_str(), stdout);
	    fflush(stdout);
	  }
	}

      pthread_cond_broadcast(&outempty);
    }
}
#endif

void handle_command(sexpr s);

void
out2client(const char *fmt, ...)
{
  va_list args;
  char buf[1024];

  va_start(args, fmt);
  std::vsnprintf(buf, sizeof(buf), fmt, args);
#ifdef OUTSERVICE
  os->push(string(buf));
#else
  {
    mutex x(&outm);
    outq.push(string(buf));
    pthread_cond_signal(&outc);
  }
#endif
  va_end(args);
}

float energy0 = 0;
float energy1 = 0;

const int MAXDELAY = 44100 * 2;

float echo_depth0;
int echo_delay0;
float echo_depth1;
int echo_delay1;

float pipe0[MAXDELAY];
float pipe1[MAXDELAY];

// inpipe is the amount of data in the pipe
// next is where to write the pipe next
int inpipe0, next0;
int inpipe1, next1;

ButterJam eq0(2, 3, 44100);
ButterJam eq1(2, 3, 44100);
bool eqp = false;
typedef ButterJam eq;

void
mix(size_t x, float *fbuf, float *fbuf1, float *fbufm, 
    float level, float level1, bool energyp = false, bool echop = false, 
    bool eqp = false)
{
  float e0 = 0, e1 = 0;

  if (eqp)
    {
      float *fbuf0e = new float[2 * x];
      float *fbuf1e = new float[2 * x];
      eq0.go(fbuf, fbuf0e, x);
      eq1.go(fbuf1, fbuf1e, x);
      memcpy(fbuf, fbuf0e, sizeof(float) * x * 2);
      memcpy(fbuf1, fbuf1e, sizeof(float) * x * 2);
      delete [] fbuf0e;
      delete [] fbuf1e;
    }

  if (echop)
    {
      //printf("foo %d %d!\n", inpipe0, echo_delay0);
      for (size_t i = 0; i < x*2; i++)
	{
	  float v0 = *fbuf++;
	  if (echo_depth0 > 0 && inpipe0 >= echo_delay0)
	    {
	      int i = next0 - echo_delay0;
	      if (i < 0) i =  MAXDELAY + i;
	      v0 += echo_depth0 * pipe0[i];
	    }
	  pipe0[next0] = v0;
	  next0 = (next0 + 1) % MAXDELAY;
	  if (inpipe0 < MAXDELAY)
	    inpipe0++;
	  v0 *= level;
	  e0 += fabs(v0);

	  float v1 = *fbuf1++;
	  if (echo_depth1 > 0 && inpipe1 >= echo_delay1)
	    {
	      int i = next1 - echo_delay1;
	      if (i < 0) i =  MAXDELAY + i;
	      v1 += echo_depth1 * pipe1[i];
	    }
	  pipe1[next1] = v1;
	  next1 = (next1 + 1) % MAXDELAY;
	  if (inpipe1 < MAXDELAY)
	    inpipe1++;
	  v1 *= level1;
	  e1 += fabs(v1);
	  
	  *fbufm++ = v0 + v1;
	}
    }
  else
    {
      for (size_t i = 0; i < x*2; i++)
	{
	  float v0 = *fbuf++;
	  v0 *= level;
	  e0 += fabs(v0);
	  float v1 = *fbuf1++;
	  v1 *= level1;
	  e1 += fabs(v1);
	  
	  *fbufm++ = v0 + v1;
	}
    }
  
  if (energyp)
    {
      energy0 = e0/(x*2);
      energy1 = e1/(x*2);
    }
}

float rate0 = 1.0, rate1 = 1.0, level0, level1;

// based on two players..this is the max of the output you'd want each to be
float level_share0 = .5;
float level_share1 = .5;

bool cue0 = false, cue1 = false;
bool cuep = false;
bool quadraticp = false;

void
spew_status()
{
}

string
scheme_escape_string(string s)
{
  string s1 = "\"";

  for (size_t i = 0; i < s.size(); i++)
    {
      switch (s[i])
	{
	case '\"':
	case '\\':
	  s1 += '\\';
	  s1 += s[i];
	  break;
	case '\n':
	  s1 += '\\';
	  s1 += 'n';
	  break;
	default:
	  s1 += s[i];
	  break;
	}
    }

  s1 += "\"";

  //cerr << s << s1 << endl;

  return s1;
}

wavfilea *player0, *player1;

/*
want to create valid wav file:
  output leading 44 bytes
  at eof, fill in header info and write it
*/

#if 0
class fileout : public ServiceQueue< pair<unsigned char *, size_t> >
{
  FILE *f;

  void service(pair<unsigned char *, size_t> p)
  {
    if (fwrite(p.first, 1, p.second, f) != p.second)
      /* how to report error? */ ;      
    delete [] p.first;
  }

  bool closingp;

public:

  class OpenError : public exception
  {
  public:
    const string file;
    const string why;
    OpenError(const char *file, const char *why)
    : file(file), why(why)
    {}
  };

  fileout(const char *file, bool appendp=false)
    : closingp(false)
  {
    f = fopen(file, (appendp ? "ab" : "wb"));
    if (!f) throw OpenError(file, strerror(errno));
  }

  write(unsigned char *p, size_t n)
  {
    if (!closingp)
      {
	unsigned char *p1 = new unsigned char[n];
	memcpy(p1, p, n);
	push(pair<unsigned char *, size_t>(p1, n));
      }
  }

  void close()
  {
    closingp = true;
    wait();
    fclose(f);
  }
};
#endif

FILE *fo;

bool shutdownp = false;

inline float
fnormalize(float f)
{
  if (f > 1)
    return 1;
  else if (f < 0)
    return 0;
  else
    return f;
}

// make this table driven?
void
handle_command(sexpr s)
{
  dprintf(10, "locking the_mutex in handle_command()\n");
  dmpmutex(&the_mutex);

  estlib::mutex x(&the_mutex);

  dprintf(10, "the_mutex locked in handle_command()\n");

  dmpmutex(&the_mutex);

  if (s.pairp())
    {
      sexpr a(s.car());
      sexpr l(s.cdr());
      
      if (sexpr::eq(a, sexpr::symbol("shutdown")))
	shutdownp = true;
      else if (sexpr::eq(a, sexpr::symbol("end-file-output")))
	{
	  if (fo) 
	    {
	      dprintf(1, "about to close fileout...\n");
	      fclose(fo);
	      dprintf(1, "fileout closed\n");
	      fo = 0;
	      out2client("(end-file-output)\n");
	    }
	  else
	    {
	      out2client("(error \"no current output file\")\n");
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("output-to-file")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      
	      if (f.stringp())
		{
		  if (fo) 
		    {
		      dprintf(1, "about to close fileout...\n");
		      fclose(fo);
		      dprintf(1, "fileout closed\n");
		      fo = 0;
		    }

		  if ((fo = fopen(string(f).c_str(), "wb")))
		    {
		      out2client("(output-to-file %s)\n", 
				 scheme_escape_string(string(f)).c_str());
		    }
		  else
		    {
		      string s = "couldn't open " + string(f)
			+ " for writing: " + string(strerror(errno));
		      out2client("(error %s)\n",
				 scheme_escape_string(s).c_str());
		    }
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("append-output-to-file")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      
	      if (f.stringp())
		{
		  if (fo) 
		    {
		      dprintf(1, "about to close fileout...\n");
		      fclose(fo);
		      dprintf(1, "fileout closed\n");
		      fo = 0;
		    }

		  if ((fo = fopen(string(f).c_str(), "ab")))
		    {
		      out2client("(append-output-to-file %s)\n", 
				 scheme_escape_string(string(f)).c_str());
		    }
		  else
		    {
		      string s = "couldn't open " + string(f)
			+ " for appending: " + string(strerror(errno));
		      out2client("(error %s)\n",
				 scheme_escape_string(s).c_str());
		    }
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("file0")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      
	      if (f.stringp())
		{
		  wavfilea *p0 = player0;

		  if (player0)
		    {
		      player0 = 0;
		      out2client("(file0-cleared)\n");
		    }
		  
		  try
		    {
		      {
			nomutex xx(&the_mutex);
			p0 = new wavfilea(f, 44100, 2, quadraticp);
		      }
		      inpipe0 = 0;
		      player0 = p0;
                      out2client("(file0 %s %f)\n", 
				 scheme_escape_string(string(f)).c_str(),
				 player0->size());
		    }
		  catch (...)
		    {
		      string s = "couldn't load " + string(f) + " on player0";
		      out2client("(error %s)\n",
				 scheme_escape_string(s).c_str());
		    }
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("file1")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      
	      if (f.stringp())
		{
		  wavfilea *p1 = player1;

		  if (player1)
		    {
		      player1 = 0;
		      out2client("(file1-cleared)\n");
		    }
		  
		  try
		    {
		      {
			nomutex xx(&the_mutex);
			p1 = new wavfilea(f, 44100, 2, quadraticp);
		      }
		      inpipe1 = 0;
		      player1 = p1;
                      out2client("(file1 %s %f)\n", 
				 scheme_escape_string(string(f)).c_str(),
				 player1->size());
		    }
		  catch (...)
		    {
		      string s = "couldn't load " + string(f) + " on player1";
		      out2client("(error %s)\n",
				 scheme_escape_string(s).c_str());
		    }
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("cue0")))
	{
	  cue0 = !cue0;
	}
      else if (sexpr::eq(a, sexpr::symbol("cue1")))
	{
	  cue1 = !cue1;
	}
      else if (sexpr::eq(a, sexpr::symbol("rate0")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		rate0 = f;
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("rate1")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		rate1 = f;
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("setloop0")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		{
		  sexpr l1(l.cdr());
		  if (l1.pairp() && l1.car().floatp())
		    {
		      if (player0)
			player0->setloop(f, l1.car());
		    }
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("setloop1")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		{
		  sexpr l1(l.cdr());
		  if (l1.pairp() && l1.car().floatp())
		    {
		      if (player1)
			player1->setloop(f, l1.car());
		    }
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("clearloop0")))
	{
	  if (player0)
	    player0->clearloop();
	}
      else if (sexpr::eq(a, sexpr::symbol("clearloop1")))
	{
	  if (player1)
	    player1->clearloop();
	}
      else if (sexpr::eq(a, sexpr::symbol("position0-delta")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		if (player0)
		  {
		    float f1 = player0->tell();
		    float f2 = float(f) + f1;
		    dprintf(1, "setting player0 to %f = %f + %f\n",
			    f2, f1, float(f));
		    player0->seek(f2);
		  }
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("position1-delta")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		if (player1)
		  {
		    float f1 = player1->tell();
		    dprintf(1, "setting player1 to %f = %f + %f\n",
			    f1 + float(f), f1, float(f));
		    player1->seek(float(f) + f1);
		  }
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("position0-get")))
	{
	  out2client("(position0 %f)\n", (player0 ? player0->tell() : 0.0));
	}
      else if (sexpr::eq(a, sexpr::symbol("position1-get")))
	{
	  out2client("(position1 %f)\n", (player1 ? player1->tell() : 0.0));
	}
      else if (sexpr::eq(a, sexpr::symbol("energy0-get")))
	{
	  out2client("(energy0 %f)\n", log(energy0 +1)/log(2));
	}
      else if (sexpr::eq(a, sexpr::symbol("energy1-get")))
	{
	  out2client("(energy1 %f)\n", log(energy1 + 1)/log(2));
	}
      else if (sexpr::eq(a, sexpr::symbol("position0")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		if (player0)
		  {
		    dprintf(1, "setting player0 to %f\n", float(f));
		    player0->seek(f);
		  }
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("position1")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		if (player1)
		  player1->seek(f);
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("eq0")))
	{
	  if (l.pairp() && l.cdr().pairp() && l.cdr().cdr().pairp())
	    {
	      sexpr fl(l.car());
	      sexpr fm(l.cdr().car());
	      sexpr fh(l.cdr().cdr().car());
	      if (fl.floatp() && fm.floatp() && fh.floatp())
		{
		  eq0.adjust(0, 0, fnormalize(fl));
		  eq0.adjust(1, 0, fnormalize(fl));
		  eq0.adjust(0, 1, fnormalize(fm));
		  eq0.adjust(1, 1, fnormalize(fm));
		  eq0.adjust(0, 2, fnormalize(fh));
		  eq0.adjust(1, 2, fnormalize(fh));
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("eq1")))
	{
	  if (l.pairp() && l.cdr().pairp() && l.cdr().cdr().pairp())
	    {
	      sexpr fl(l.car());
	      sexpr fm(l.cdr().car());
	      sexpr fh(l.cdr().cdr().car());
	      if (fl.floatp() && fm.floatp() && fh.floatp())
		{
		  eq1.adjust(0, 0, fnormalize(fl));
		  eq1.adjust(1, 0, fnormalize(fl));
		  eq1.adjust(0, 1, fnormalize(fm));
		  eq1.adjust(1, 1, fnormalize(fm));
		  eq1.adjust(0, 2, fnormalize(fh));
		  eq1.adjust(1, 2, fnormalize(fh));
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("echo0")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      sexpr fx(l.cdr().car());
	      if (f.floatp() && fx.floatp())
		{
		  echo_delay0 = 2 * (int) floor(fnormalize(f) * 44100);
		  echo_depth0 =  fnormalize(fx);
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("echo1")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      sexpr fx(l.cdr().car());
	      if (f.floatp() && fx.floatp())
		{
		  echo_delay1 = 2 * (int) floor(fnormalize(f) * 44100);
		  echo_depth1 =  fnormalize(fx);
		}
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("level0")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		level0 = fnormalize(f);
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("level1")))
	{
	  if (l.pairp())
	    {
	      sexpr f(l.car());
	      if (f.floatp())
		level1 = fnormalize(f);
	    }
	}
      else if (sexpr::eq(a, sexpr::symbol("xfade")))
	{
	  if (l.pairp() && l.cdr().pairp()
	      && l.car().floatp() && l.cdr().car().floatp())
	    {
	      float f0 = l.car();
	      float f1 = l.cdr().car();
	      // !!! should really check that they sum to 1.0
	      level_share0 = fnormalize(f0);
	      level_share1 = fnormalize(f1);
	    }
	  else
	    out2client("(error \"malformed xfade message\")\n");
	}
      else if (sexpr::eq(a, sexpr::symbol("status")))
	spew_status();
    }
}

unsigned doion;

void *
doio(void *v)
{
  dprintf(1, "starting doio()\n");

  while (true)
    {
      sexpr s;

      cin >> s;

      dprintf(100, "doio read sexpr number %u\n", doion++);

      /* NB: this ensures we don't go into a tight loop of a
	non-blocking select() followed by a non-blocking read..this is
	critical for a real-time program!  */
      if (s.eof_objectp())
	{
	  dprintf(10, "locking the_mutex for eof in doio()\n");
	  {
	    dmpmutex(&the_mutex);
	    mutex x(&the_mutex);
	    dprintf(10, "the_mutex locked in doio()\n");
	    dmpmutex(&the_mutex);
	    shutdownp = true;
	  }
	  return 0;
	}

      handle_command(s);
    }
}

const char *cue_channel = 0;

bool pci128_lossage = false;
int fragment = 0x00070009;

void
pci128fix(float *f, size_t n)
{
  for (size_t i = 0; i < n; i++)
    {
      f[2*i+1] = (f[2*i] + f[2*i+1])/2;
      f[2*i] = 0;
    }
}

enum
{
  OPT_LF, OPT_MF, OPT_MBW, OPT_HF
};

float lf = 250, mf = 1250, mbw = 2500, hf = 2500;

struct option options[] =
{{"fragment", 1, 0, 'f'},
 {"pci128-lossage", 0, 0, 'p'},
 {"cue-channel", 1, 0, 'c'},
 {"debug", 1, 0, 'd'},
 {"quadratic", 0, 0, 'q'},
 {"eq", 0, 0, 'e'},
 {"lf", 1, 0, OPT_LF},
 {"mf", 1, 0, OPT_MF},
 {"mbw", 1, 0, OPT_MBW},
 {"hf", 1, 0, OPT_HF},
 {0, 0, 0, 0}};

#ifdef __linux__
extern "C" int fileno(FILE *);
#endif

void
set_eq_params(eq &e, float lf, float mf, float mbw, float hf)
{
  e.adjust(0, 0, 1.0, ButterJam::LOWPASS, lf);
  e.adjust(1, 0, 1.0, ButterJam::LOWPASS, lf);
  e.adjust(0, 1, 1.0, ButterJam::BANDPASS, mf, mbw);
  e.adjust(1, 1, 1.0, ButterJam::BANDPASS, mf, mbw);
  e.adjust(0, 2, 1.0, ButterJam::HIGHPASS, hf);
  e.adjust(1, 2, 1.0, ButterJam::HIGHPASS, hf);
}

// descriptor for control input thread
pthread_t iot;

bool clerror = false;

void
handle_float_opt(const char *opt, const char *val, float *loc)
{
  float f;
  char c;
	  
  if (sscanf(val, "%f%c", loc, &c) == 1)
    ;
  else
    {
      dprintf(0, "%s: malformed argument to %s: %s\n",
	      opt, progname, val);
      clerror = true;
    }
}

pthread_t pingt;

unsigned pingn;

void *
pinger(void *v)
{
  while (true)
    {
      char buf[100];
      sprintf(buf, "%u:", pingn++);
      write(2, buf, strlen(buf));
    }
  return 0;
}

// mixd
int
main1(int argc, char **argv)
{
  int idx;

  int rc;

  //os = new outservice;
  //is = new interpretservice;

  //out2client(string("(oola-mixd)\n"));

  while ((rc = getopt_long(argc, argv, "d:c:f:pq", options, &idx)) != EOF)
    switch (rc)
      {
      case ':': 
      case '?': clerror = true; break;
      case 'c': 
	cue_channel = optarg; 
	break;
      case 'e': 
	eqp = true; 
	break;
      case 'p': 
	pci128_lossage = true; 
	break;
      case 'q':
	quadraticp = true; 
	break;
      case OPT_LF: 
	handle_float_opt("--lf", optarg, &lf);
	break;
      case OPT_MF: 
	handle_float_opt("--mf", optarg, &mf);
	break;
      case OPT_MBW: 
	handle_float_opt("--mbw", optarg, &mbw);
	break;
      case OPT_HF: 
	handle_float_opt("--hf", optarg, &hf);
	break;
      case 'f': 
	{
	  int i;
	  char c;
	  
	  if (sscanf(optarg, "%x%c", &i, &c) == 1)
	    {
	      i &= 0xFFFFFFFF;
	      fragment = i;
	    }
	  else
	    {
	      dprintf(0, "%s: malformed argument to --fragment: %s\n",
		      progname, optarg);
	      clerror = true;
	    }
	break;
	}
      case 'd': 
	{
	  int i;
	  char c;
	  
	  if (sscanf(optarg, "%d%c", &i, &c) == 1)
	    {
	      debuglevel = i;
	      dprintf(0, "debuglevel = %d\n", debuglevel);
	    }
	  else
	    {
	      dprintf(0, "%s: malformed argument to --debug: %s\n",
		      progname, optarg);
	      clerror = true;
	    }
	break;
	}
      }

  if (clerror)
    return 1;

  if (eqp)
    {
      set_eq_params(eq0, lf, mf, mbw, hf);
      set_eq_params(eq1, lf, mf, mbw, hf);
    }

  dprintf(1, "cue channel: %s\n", (cue_channel ? cue_channel : "<none>"));
  dprintf(1, "pci128-lossage = %s\n", (pci128_lossage ? "yes" : "no"));
  dprintf(1, "quadratic = %s\n", (quadraticp ? "on" : "off"));
  dprintf(1, "fragment = 0x%8.8lx\n", fragment);

  if (geteuid() == 0)
    {
      struct sched_param sp;
      sp.sched_priority = 50;
      if (sched_setscheduler(0, SCHED_FIFO, &sp) == -1)
	{
	  dprintf(1, "couldn't set real-time scheduling: %s\n",
		  strerror(errno));
	  return 1;
	}
      else
	dprintf(1, "process set to real-time priority %d\n", 
		sp.sched_priority);

#ifdef _POSIX_MEMLOCK
      if (mlockall(MCL_CURRENT | MCL_FUTURE) == -1)
	dprintf(1, "problem locking memory: %s\n", strerror(errno));
      else
	dprintf(1, "current and future memory locked\n");
#endif
    }

  soundout *o = mydsp(fragment);
  if (cue_channel && (cuep = cue_try(cue_channel, fragment)))
    dprintf(1, "cue channel on\n");

  float *fbuf0 = new float [o->getbufsiz()/2];
  float *fbuf1 = new float [o->getbufsiz()/2];
  float *fbufm = new float [o->getbufsiz()/2];
  float *fbufmc = new float [o->getbufsiz()/2];
  float *fbufz = new float [o->getbufsiz()/2];

  for (size_t i = 0; i < o->getbufsiz()/2; i++)
    fbufz[i] = 0;

  level0 = 1.0;
  level1 = 1.0;

  size_t n = o->getbufsiz()/4;

  // init control input thread
  pthread_attr_t att;
  pthread_attr_init(&att);
  if (pthread_attr_setinheritsched(&att, PTHREAD_INHERIT_SCHED) == -1)
    {
      dprintf(1, 
	      "couldn't do pthread_attr_setinheritsched(&att, PTHREAD_INHERIT_SCHED): %s\n", 
	      strerror(errno));
    }

  if (pthread_create(&iot, &att, doio, 0) == -1)
    {
      dprintf(0, "couldn't create control input thread: %s\n", 
	      strerror(errno));
      return 1;
    }

#ifdef OUTSERVICE
  os = new outservice;
#else
  pthread_mutex_init(&outm, 0);
  pthread_cond_init(&outc, 0);
  pthread_cond_init(&outempty, 0);
  if (pthread_create(&outt, &att, outtask, 0) == -1)
    {
      dprintf(0, "couldn't create control output thread: %s\n", 
	      strerror(errno));
      return 1;
    }
#endif

  dprintf(10, "locking the_mutex in main1()\n");
  dmpmutex(&the_mutex);
  pthread_mutex_lock(&the_mutex);
  dmpmutex(&the_mutex);
  dprintf(10, "the_mutex locked in main1()\n");

  while (true)
    {
      if (shutdownp)
	throw EndApp();

      bool p0 = player0 && player0->goodtogetp();
      bool p1 = player1 && player1->goodtogetp();

      if (p0)
	player0->get(fbuf0, n, rate0);
      if (p1)
	player1->get(fbuf1, n, rate1);

      mix(n, 
	  (p0 ? fbuf0 : fbufz), 
	  (p1 ? fbuf1 : fbufz), 
	  fbufm, 
	  level0 * level_share0,
	  level1 * level_share1,
	  true,
	  true,
	  eqp);

      {
	// unlock mutex during output
	dprintf(100, "unlocking the_mutex during sound output\n");
	nomutex x(&the_mutex);
	dprintf(100, "the_mutex unlocked during sound output\n");
	o->put(fbufm, n);
      }

      dprintf(100, "the_mutex relocked after sound output\n");

      // do output to file if requested
      if (fo)
	{
	  unsigned char *buf1 = new unsigned char[4 * n];
	  // ack..this conversion duplicates work
	  convert2(fbufm, n, buf1);
	  // NB..give this output its own thread
	  fwrite(buf1, 1, 4*n, fo);
	}
      
      if (cuep)
	{
	  mix(n, 
	      (p0 ? fbuf0 : fbufz), 
	      (p1 ? fbuf1 : fbufz), 
	      fbufmc, 
	      (cue0 ? .5 : 0),
	      (cue1 ? .5 : 0));
	  
	  if (pci128_lossage)
	    pci128fix(fbufmc, n);
	  
	  {
	    // unlock mutex during output
	    nomutex x(&the_mutex);
	    cue_put(fbufmc, n);
	  }
	}
    }
  
  return 0;
}
  
int
main(int argc, char **argv)
{
  int rc = 0;

  fflush(stdout);

  try
    {
      rc = main1(argc, argv);
    }
  catch (bad_alloc)
    {
      out2client("(error \"out of memory\")\n");
      rc = 1;
    }
  catch (EndApp)
    {
      out2client("(shutting-down)\n");
      if (fo) fclose(fo);
      rc = 0;
    }
  catch (exception &e)
    {
      out2client("(error \"unhandled exception %s\")\n", e.what());
      rc = 1;
    }
  catch (...)
    {
      out2client("(error \"Unknown unhandled exception!\")\n");
      rc = 1;
    }
      
  // make sure output to client has drained
#ifdef OUTSERVICE
  os->wait();
#else
  {
    mutex x(&outm);
    if (outq.size() != 0)
      pthread_cond_wait(&outc, &outm);
  }
#endif

  return rc;
}
