/*
    This file is part of the oolaboola digital dj program

    Copyright (C) 1999  Eric S. Tiedemann

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <cstddef>
#include <exception>

using namespace std;

class soundfile
{
  struct samppair
  {
    float left, right;
    
    samppair(float l, float r)
      : left(l), right(r)
      {}
    
    samppair()
      : left(0), right(0)
      {}
  };
  
public:

  class UnexpectedEOF : public exception {};
  class IllFormatted : public exception {};
  class OpenError : public std::exception
    {
      char *what() { return "soundfile::OpenError"; }
    };

  virtual ~soundfile();

  virtual size_t get(float *fbuf, size_t nsamps, float resample = 1.0) = 0;
  samppair interp(float factor, samppair a, samppair b);
  virtual float tell() = 0;
  virtual void seek(float) = 0;
};

inline soundfile::samppair
soundfile::interp(float factor, samppair a, samppair b)
{
  return samppair(a.left * (1 - factor) + b.left * factor,
		  a.right * (1 - factor) + b.right * factor);
}
